# -*- encoding: utf-8 -*-
import base64
import binascii
from collections import defaultdict, namedtuple
from functools import lru_cache
import logging
import os
import random
import re
import requests
import datetime
import time
import werkzeug.urls

from odoo import SUPERUSER_ID, api, fields, models
from odoo.osv import expression
from odoo.tools import ormcache

from . import certtool

from odoo.addons.saas_worker.auth_util import SaasAuth
from odoo.addons.saas_worker.const import (
    DNS_MAX_LENGTH, DOMAIN_SRV, IOT_DOMAIN, LOCAL,
    PAAS_DOMAINS, SAAS_BASE_URL, SCHEME, WEBSITE
)
from odoo.addons.saas_worker.util import str2bool, humanize_version
from odoo.exceptions import UserError
DOMAIN_SAAS = SAAS_BASE_URL.split(':')[0]  # Strip the port

_logger = logging.getLogger(__name__)


class AutoVacuum(models.AbstractModel):
    _inherit = 'ir.autovacuum'

    @api.model
    def power_on(self, *args, **kwargs):
        self.env['ssl.certificate']._cert_gc()
        return super(AutoVacuum, self).power_on(*args, **kwargs)


class BackupGroup(models.Model):
    _name = "backup.group"
    _description = "Backup group"
    _sql_constraints = [
        ('name_uniq', 'unique(name)', 'Name must be unique')
    ]

    name = fields.Char(required=True)
    auto_config = fields.Boolean(
        "Auto Config",
        default=False,
        help="Can be automatically set at server creation",
    )
    mode = fields.Char(
        "Mode",
        help="Server modes accepted for auto attribution. Accept any mode if empty.",
    )


class SSLCertificate(models.Model):
    _name = "ssl.certificate"
    _description = "SSL Certificate"
    _order = 'subject_cn'
    _rec_name = 'handle'

    def _default_expire_date(self):
        """ Let the record expire before we attempt the renewal """
        d = datetime.datetime.now() + datetime.timedelta(days=+3)
        return fields.Datetime.to_string(d)

    handle = fields.Char("Client Identifier", required=True)
    subject_cn = fields.Char("Certificate Subject Common Name (CN)", required=True, index=True)
    subject_alt_name = fields.Char("Certificate Subject Alternative Name")
    not_valid_before = fields.Datetime("Not Valid Before Date")
    not_valid_after = fields.Datetime("Not Valid After Date")
    private_key_pem = fields.Binary("Private Key", attachment=False)
    x509_pem = fields.Binary("X509 Certificate", attachment=False)
    origin = fields.Char('Origin')
    http_01_validation = fields.Char('Validation Token to be served on the well known URL')
    http_01_validation_path = fields.Char('Specific path of the well known URL', index=True)
    expire_date = fields.Datetime("Expiry Date", default=_default_expire_date, index=True)
    last_auth_failure = fields.Datetime("Last Authorization Failure")
    last_auth_error = fields.Char("Last Error Message")

    _sql_constraints = [
        ('handle_uniq', 'unique(handle)', 'Handle must be unique.'),
        ('subject_cn_uniq', 'unique(subject_cn)', 'Subject CN must be unique.')
    ]

    def _auto_init(self):
        res = super(SSLCertificate, self)._auto_init()
        index_queries = """
            CREATE INDEX IF NOT EXISTS ssl_certificate_origin_handle_index
            ON ssl_certificate (origin, handle);
            CREATE INDEX IF NOT EXISTS ssl_certificate_origin_expire_date_index
            ON ssl_certificate (origin, expire_date);
        """
        self._cr.execute(index_queries)
        return res

    def _cert_domains(self):
        self.ensure_one()
        alt_names = self.subject_alt_name.split() if self.subject_alt_name else []
        domains = sorted(list(set([self.subject_cn] + alt_names)))
        return domains

    def _issue_and_write(self, domains, chall_type):
        self.ensure_one()
        certtool_env = os.getenv("CERTTOOL_ENV", "staging")
        pkey_pem, fullchain_pem, x509_info = certtool.auth_and_issue(
            domains=domains,
            chall_type=chall_type,
            email="certs-saas@odoo.com",
            env=certtool_env,
            cr=self.env.cr)
        self.write({
            'subject_cn': x509_info['subject_cn'],
            'subject_alt_name': ' '.join(x509_info['subject_alt_name']),
            'not_valid_before': x509_info['not_valid_before'],
            'not_valid_after': x509_info['not_valid_after'],
            'private_key_pem': pkey_pem,
            'x509_pem': fullchain_pem,
            'last_auth_failure': False,
            'last_auth_error': '',
        })
        log_msg = ("x509: new certificate: subject_cn: %s, "
                   "subject_alt_name: %s, "
                   "not_valid_after: %s")
        _logger.info(
            log_msg, x509_info['subject_cn'], x509_info['subject_alt_name'],
            x509_info['not_valid_after'])

    def _issue_certificates(self):
        chall_type_map = {
            'internal': 'dns-01',
            'iot': 'dns-01',
            'saas': 'http-01',
        }
        # Max retry frequency: once every 60min
        # Use tanh for a simple capped non-linear backoff:
        # Days -> Days: 5 * (tanh(x - 4) + 1)
        # cooldown capped at 10d, inflection at 4d, f(0) ~ 0
        # stop trying after 6 weeks
        certs_query = """
            WITH to_issue AS (
                SELECT
                    c.id,
                    COALESCE(c.last_auth_failure, to_timestamp(0)) AS last_attempt,
                    CASE
                    WHEN not_valid_after IS NOT NULL THEN
                        (EXTRACT (EPOCH FROM now() AT TIME ZONE 'UTC') -
                         EXTRACT (EPOCH FROM c.not_valid_after - interval '30 days'))
                         / (3600 * 24)
                    WHEN last_auth_failure IS NULL THEN
                        0
                    ELSE
                        (EXTRACT(EPOCH FROM now() AT TIME ZONE 'UTC') -
                         EXTRACT(EPOCH FROM c.create_date))
                         / (3600 * 24)
                    END AS authf_days
                FROM  ssl_certificate AS c
                WHERE (c.x509_pem IS NULL OR c.not_valid_after <=
                       now() AT TIME ZONE 'UTC' + interval '30 days')
            )
            SELECT i.id
            FROM   to_issue AS i
            WHERE  i.last_attempt <= now() AT TIME ZONE 'UTC' - interval '1 day'
                   * 5 * (tanh(authf_days - 4) + 1)
            AND    (i.last_attempt <= now() AT TIME ZONE 'UTC' - interval '60 minutes')
            AND    i.authf_days <= 60
            LIMIT 50
        """
        now = datetime.datetime.now()
        self.env.cr.execute(certs_query)
        certs_to_issue = self.browse([r[0] for r in self.env.cr.fetchall()])
        for c in certs_to_issue:
            domains = c._cert_domains()
            chall_type = chall_type_map.get(c.origin, 'dns-01')
            try:
                c._issue_and_write(domains, chall_type)
            except Exception as e:
                c.last_auth_failure = now
                c.last_auth_error = e
                _logger.warning("Failed to issue cert for domains %s. Error: %s",
                             domains, e)

    def _update_expiry(self):
        now = datetime.datetime.now()
        long_expiry_low_bound = now + datetime.timedelta(days=+30)
        long_expiry_high_bound = now + datetime.timedelta(days=+45)
        short_expiry_low_bound = now + datetime.timedelta(days=+1)
        short_expiry_high_bound = now + datetime.timedelta(days=+3)

        for c in self:
            if c.not_valid_after and c.not_valid_after > now:
                expiry_low_bound = long_expiry_low_bound
                expiry_high_bound = long_expiry_high_bound
            else:
                expiry_low_bound = short_expiry_low_bound
                expiry_high_bound = short_expiry_high_bound
            if c.expire_date < expiry_low_bound:
                c.write({'expire_date': expiry_high_bound})

    def _require_iot_certs(self, required_handles):
        certs = self.search(
            [('handle', 'in', required_handles),
             ('origin', '=', 'iot')])
        existing_handles = certs.mapped('handle')
        diff_handles = list(set(required_handles) - set(existing_handles))
        for h in diff_handles:
            subject_cn = '*.{}.{}'.format(h, IOT_DOMAIN)
            subject_alt_name = ' '.join([subject_cn, subject_cn[2:]])
            rec = self.create({
                'handle': h,
                'origin': 'iot',
                'subject_cn': subject_cn,
                'subject_alt_name': subject_alt_name,
            })
            certs |= rec
            _logger.info('created new iot cert. handle: %s', h)
        certs._update_expiry()
        return certs

    def _export_iot_cert(self, handle):
        cert = self.search([('handle', '=', handle), ('origin', '=', 'iot')])
        if not cert:
            # Issue on the fly
            otf_cert = self._require_iot_certs([handle])
            domains = otf_cert._cert_domains()
            try:
                otf_cert._issue_and_write(domains, 'dns-01')
            except Exception as e:
                _logger.warning("Failed to issue cert for domains %s. Error: %s",
                             domains, e)
                return None, None, None
            cert = otf_cert
        x509_info = {
            'subject_cn': cert.subject_cn,
        }
        return cert.private_key_pem, cert.x509_pem, x509_info

    def _require_certs(self, domains, origin='saas', wildcard=False):
        """ domains: required saas/internal domains """
        certs = self.search(
            [('handle', 'in', domains), ('origin', '=', origin)])
        new_handles = list(set(domains) - set(certs.mapped('handle')))
        for h in new_handles:
            rec_data = {
                'handle': h,
                'origin': origin,
            }
            if origin != 'saas' and wildcard:
                subject_cn = '*.' + h
                rec_data.update({
                    'subject_cn': subject_cn,
                    'subject_alt_name': ' '.join([subject_cn, subject_cn[2:]]),
                })
            else:
                subject_cn = h
                rec_data.update({
                    'subject_cn': subject_cn,
                })
            rec = self.create(rec_data)
            certs |= rec
            _logger.info('created new saas cert. handle: %s', h)
        certs._update_expiry()
        return certs

    def _export_certs(self, handles, origin='saas'):
        certs = self.search(
            [('handle', 'in', handles),
             ('origin', '=', origin),
             ('not_valid_after', '!=', False)])
        certs_info = []
        for cert in certs:
            certs_info.append({
                'domain': cert.subject_cn,
                'private_key_pem': cert.private_key_pem,
                'x509_pem': cert.x509_pem,
            })
        return certs_info

    def _refresh(self, handles, origin='saas'):
        """In case of previous errors, clears the Last Authorization Failure
        in order to issue again with less wait time. PEM files are also
        cleared in case we want to force the generation of new certificates."""
        _logger.info("refresh certs: origin=%s handles=%s", origin, handles)
        certs = self.search([
            ('handle', 'in', handles),
            ('origin', '=', origin),
        ])
        certs.write({
            'not_valid_before': False,
            'not_valid_after': False,
            'last_auth_failure': False,
            'last_auth_error': '',
            'x509_pem': False,
            'private_key_pem': False,
        })
        return certs

    @api.model
    def _cert_gc(self):
        """ gc iot and saas certificates """
        now = datetime.datetime.now()
        expired_certs = self.search(
            [('origin', 'in', ('iot', 'saas')),
             ('expire_date', '<', fields.Datetime.to_string(now))])
        _logger.info('gc certs. handles: %s', expired_certs.mapped('handle'))
        expired_certs.unlink()


class SaasDC(models.Model):
    _name = "saas.dc"
    _description = "SaaS DC"
    _order = "provider, name"

    provider = fields.Char(required=True)
    name = fields.Char(required=True)

    @api.multi
    def name_get(self):
        return [(dc.id, '%s - %s' % (dc.provider, dc.name))
                for dc in self]
        # Uncomment me when in 17.0 please :-)
        # for dc in self:
        #     dc.display_name = f'{dc.provider} - {dc.name}'

    @api.model
    def name_search(self, name, args=None, operator='ilike', limit=100):
        if name and operator not in expression.NEGATIVE_TERM_OPERATORS:
            domain = ['|', ('name', operator, name), ('provider', operator, name)]
            data_centers = self.search(domain + (args or []), limit=limit)
            return data_centers.name_get()
        return super(SaasDC, self).name_search(name, args, operator, limit=limit)


class SaasDNS(models.Model):
    _name = "saas.dns"
    _description = "SaaS DNS"
    _table = "records"  # used by powerdns
    _order = 'name'

    name = fields.Char('Name', required=True)
    domain_id = fields.Integer('domain_id', required=True, default=1)
    # 'domain_id': fields.many2one('saas.domain', 'domain'),
    type = fields.Char('Type', required=True, default='A')
    content = fields.Char('Content', required=True)
    ttl = fields.Integer('TTL', required=True, default=600)
    prio = fields.Integer('prio', required=True, default=0)
    change_date = fields.Integer('Change Time', compute='_compute_change_time', store=True)
    origin = fields.Char('Origin')
    # Used by PowerDNS, default is important!
    disabled = fields.Boolean('Disabled', required=True, default=False)
    ordername = fields.Char('Ordername')
    auth = fields.Boolean('Auth', required=True, default=True)

    # Used to auto-increment pdns zone serial when SOA has serial=0
    def _compute_change_time(self):
        timestamp = time.time()
        for server in self:
            server.change_date = timestamp

    _DNS = namedtuple('DNS', 'name type content ttl prio origin')

    @api.multi
    def unlink(self):
        reason = self._context.get('reason', 'manual')
        log = _logger.getChild('dns').info
        dns_data = self.read(self._DNS._fields)
        result = super(SaasDNS, self).unlink()
        for data in dns_data:
            data_id = data.pop('id')
            log('remove dns#%s (reason: %s): %r', data_id, reason, self._DNS(**data))
        return result

    # -------------------------------------------------
    # Cron methods
    # -------------------------------------------------

    @api.model
    def _dns_gc(self):
        """
        Garbage collect DNS names.

        Search for orphaned DNS records and delete them. This part can be done
        outside the stats_update critical section.
        """
        # gc orphan TXT records
        orphan_records_query = """
            WITH valid_records AS (
                SELECT name
                  FROM records
                 WHERE origin IN ('saas', 'saas_booking', 'paas') AND
                       type IN ('A', 'MX')
              GROUP BY name
            )
                SELECT array_agg(orphan_records.id)
                  FROM records AS orphan_records
             LEFT JOIN valid_records USING (name)
                 WHERE orphan_records.type in ('TXT') AND
                       orphan_records.origin in ('saas', 'saas_booking', 'paas') AND
                       array_length(regexp_split_to_array(split_part(orphan_records.name, %s, 1), '\\.'), 1) = 1 AND
                       valid_records.name IS NULL
        """
        self.env.cr.execute(orphan_records_query, (f".{SAAS_BASE_URL}",))
        dns_ids = self.env.cr.fetchone()[0]
        dns = self.env["saas.dns"].browse(dns_ids)
        dns.with_context(reason="dropped orphan record").unlink()

    # -------------------------------------------------
    # Actions
    # -------------------------------------------------

    def action_open_dns_rw(self):
        action = {
            "name": "Edit SaaS DNS",
            "type": "ir.actions.act_window",
            "res_model": "saas.dns",
            "views": [[self.env.ref('saas_master.saas_dns_rw_form').id, "form"]],
            "res_id": self.id,
        }
        return action


class SaasServer(models.Model):
    _name = "saas.server"
    _description = "SaaS Server"
    _order = 'name'

    name = fields.Char('Name')
    public_ip = fields.Char('Public IP', required=True)
    public_ipv6 = fields.Char('Public IPv6')
    mode = fields.Char('Mode', help="either: demo, trial, paid, bak, slave")
    profiles = fields.Char(
        'Profiles',
        compute='_compute_profiles',
        store=True,
        help='Server installation profile: odoo, pg or basic',
    )
    master = fields.Char('Master', help="master server")
    hosting = fields.Char('Hosting')
    dc_id = fields.Many2one('saas.dc', 'Data Center')
    dc_provider = fields.Char(related='dc_id.provider', store=True, readonly=True)
    backup_group_id = fields.Many2one('backup.group', 'Backup Group')
    public_dns = fields.Char('Public dns')
    public_ssh_key = fields.Char('Public SSH key')
    http_shared_secret = fields.Char('HTTP shared secret', required=True, default=lambda x: base64.b64encode(os.urandom(32)))
    lastping = fields.Datetime('Lastping')
    private_ip = fields.Char('Private IP')
    private_dns = fields.Char('Private dns')
    password = fields.Char('Password')  # TODO: delete
    stat_cpu = fields.Integer('CPU')
    stat_db_all = fields.Integer('Dbs')
    stat_db_trial = fields.Integer('Trials')
    stat_db_pay = fields.Integer('Paying')
    stat_disk_free = fields.Integer('Free Space (GiB)')
    stat_disk_total = fields.Integer('Total Space (GiB)')
    stat_loadavg = fields.Char('Load')
    stat_ram = fields.Integer('RAM (GiB)')
    stat_revs = fields.Char('Revs')
    last_branch = fields.Char('Last branch')
    backuped = fields.Boolean('Backuped', default=True)  # deprecated
    mode_munin_config = fields.Text('Munin Config')
    cname_list = fields.Char('CNAME list', compute='_compute_cname_list')
    replicas = fields.Many2many(
        string='Replicas',
        comodel_name='saas.server',
        relation='saas_server_replica_rel',
        column1='primary_id',
        column2='replica_id',
        copy=False,
    )
    primaries = fields.Many2many(
        string='Primaries',
        comodel_name='saas.server',
        relation='saas_server_replica_rel',
        column1='replica_id',
        column2='primary_id',
        copy=False,
    )
    rack = fields.Char('Rack')
    machine_family = fields.Char('Machine Family')
    tenant = fields.Char(
        'Tenant',
        compute='_compute_tenant',
        store=True,
        help='Server tenant',
    )

    _sql_constraints = [
        ('name_uniq', 'unique(name)', 'Name must be unique')
    ]

    @api.multi
    def _compute_cname_list(self):
        for server in self:
            dns_entries = self.env['saas.dns'].search([('type', '=', 'CNAME'), ('content', '=', server.name)])
            server.cname_list = ' '.join(dns_entries.mapped('name'))

    @api.depends('mode')
    def _compute_profiles(self):
        default_profiles = ["pg", "odoo"]
        profile_modes = {
            "basic": {
                "bak",
                "build",
                "consul",
                "dbex",
                "doc2",
                "filestore",
                "git",
                "grafana",
                "jitsi-meet",
                "jitsi-vb",
                "lb",
                "loki",
                "mail",
                "migbot",
                "mimir",
                "munin",
                "mx",
                "mq",
                "prom",
                "sfu",
                "upgrade_master",
                "vault",
                "wheatbot",
            },
            "pg": {
                "analytics",
                "dns-iot",
                "pg_standby",
            }
        }
        for server in self:
            # PaaS servers have separate management
            if server.mode and "paas" in server.mode:
                server.profiles = ""
                continue
            if server.mode:
                modes = server.mode.split()
            else:
                modes = []
            host_profiles = []
            for saas_profile, saas_modes in profile_modes.items():
                if saas_modes & set(modes):
                    host_profiles.append(saas_profile)
            if not host_profiles:
                # default profiles
                host_profiles = default_profiles
            server.profiles = " ".join(host_profiles)

    @api.depends("mode", "name", "backup_group_id.mode")
    def _compute_tenant(self):
        default_tenant = "internal"
        tenant_modes = {
            "odoofin": {
                "fin",
            },
            "odoosh": {
                "paas_master",
                "paas_tech",
                "paas_worker",
            },
            "saas": {
                "trial",
                "pg_standby",
            },
            "upgrade": {
                "upgrade",
                "upgrade_ci",
                "upgrade_master",
            },
        }
        for server in self:
            server.tenant = default_tenant
            if not server.mode:
                continue
            modes = server.mode.split()

            if "bak" in modes:
                # infer from naming convention
                server.tenant = "internal" if server.name.startswith("bak-in") else "saas"
                if server.backup_group_id.mode:
                    # infer from backup group auto-configuration
                    bg_modes = server.backup_group_id.mode.split()
                    modes.extend(bg_modes)

            for tenant, mode in tenant_modes.items():
                if mode & set(modes):
                    server.tenant = tenant

    @api.model
    def create(self, values):
        server = super(SaasServer, self).create(values)
        name = values.get('name')
        name = name or ('srv%d.%s' % (server.id, DOMAIN_SRV))
        name = name.lower()

        server.write({'name': name})
        self.env['saas.dns'].create({
            'name': name,
            'content': values['public_ip'],
            'origin': 'server'
        })
        return server

    @api.multi
    def write(self, values):
        if values.get('name'):
            values['name'] = values['name'].lower()
        result = super(SaasServer, self).write(values)
        if 'name' in values:
            for server in self:
                domain = [
                    ('content', '=', server.public_ip),
                    ('origin', '=', 'server')
                ]
                self.env['saas.dns'].search(domain).write({'name': values['name']})
        return result

    def _prepare_replication_vals(self, replication_mode, master=None, replicas=None):
        master_server = None
        if replication_mode in ('dedicated', 'hybrid') and master:
            master_server = self.search([('name', '=', master)])
            if not master_server:
                raise ValueError('Master server does not exist: %s', master)

        replica_servers = self.env['saas.server']
        if replication_mode in ('shared', 'hybrid') and replicas:
            hosts = ['{}.{}'.format(r, DOMAIN_SAAS) for r in replicas]
            replica_servers = self.search([('name', 'in', hosts), ('mode', '=', 'pg_standby')])
            if len(replicas) != len(replica_servers):
                raise ValueError('One or several replica servers do not exist: %s' % replicas)

        vals = {}

        if replication_mode in ('dedicated', 'hybrid') and master:
            # The 'master' field must always be rewritten to clean-up previous values
            vals['master'] = master
            vals['primaries'] = [(6, 0, master_server.ids)]
        elif replication_mode in ('shared', 'hybrid') and replicas:
            vals['master'] = None
            vals['replicas'] = [(6, 0, replica_servers.ids)]
        else:
            vals['master'] = None
            vals['primaries'] = None
            vals['replicas'] = None

        return vals

    # -------------------------------------------------
    # Trial, database and DNS operations
    # -------------------------------------------------

    def demo(self, user, key, module=None, hosting='eu', redirect=None):
        servers = self.search([('mode', 'like', 'demo'), ('hosting', '=ilike', hosting)])
        if not servers:
            servers = self.search([('mode', 'like', 'demo')])
        if not servers:
            return None
        server = random.choice(servers)

        params = {
            'user': user,
            'key': key,
            'module': module,
            'redirect': redirect,
        }
        url = '{0}://{1}/saas_worker/demo?{2}'.format(SCHEME, server.name, werkzeug.urls.url_encode(params))
        return url

    @api.model
    def trial_check(self, dbname, spam_check=True):
        """ validate a dbname """
        dbname = dbname.lower()
        if dbname == 'meta':
            raise ValueError('Reserved name')
        # we limit the max db name to DNS_MAX_LENGTH characters, postgres db name is limit to 64
        # and for special cases (db creation, deletion, etc ...) we move the dbs with their name prefixed
        if len(dbname) > DNS_MAX_LENGTH:
            raise Exception('Maximum name length is {} for DB {}'.format(DNS_MAX_LENGTH, dbname))

        if not re.match("^[a-z0-9][a-z0-9-]{2,}[a-z0-9]$", dbname):
            raise Exception('Invalid name')

        hostname = '%s.%s' % (dbname, DOMAIN_SAAS)
        domain = [
            ('name', '=', hostname),
        ]

        db_count = self.env['saas.dns'].search_count(domain)

        if db_count:
            raise ValueError('domain already taken')

        if spam_check:
            try:
                valid_spam_check = self._trial_check_spam(dbname)
            except:
                valid_spam_check = True

            if not valid_spam_check:
                raise ValueError('domain already taken')

    @ormcache('dbname')
    def _trial_check_spam(self, dbname):
        """ Check if the name is not blacklisted by trial spam rules on odoo.com
            Return: True if it pass the spam check, False if not
        """
        payload = {'params':{
                'name': dbname
            }
        }
        result = requests.get(WEBSITE + '/oe-trial-rules/name', json=payload, auth=SaasAuth(), timeout=5)
        result.raise_for_status()
        json_res = result.json()
        return json_res['result']['valid']

    @api.model
    def domain_check(self, domain, origin, check_exists=True):
        domain = domain.lower()
        try:
            dbname, subdomain = domain.split('.', 1)
        except ValueError:
            raise Exception('invalid name')
        if origin == 'paas':
            if subdomain not in PAAS_DOMAINS:
                raise Exception('invalid subdomain')
        elif subdomain != DOMAIN_SAAS:
            raise Exception('invalid subdomain')
        if not re.match(r"^[a-z0-9][a-z0-9\-]{2,}[a-z0-9]$", dbname):
            raise Exception('Invalid name')
        if check_exists:
            exists = self.env['saas.dns'].search_count([('name', '=', domain)])
            if exists:
                raise Exception('domain already taken')
        return True

    @api.model
    def trial_find_dbname(self, dbname):
        """ find a free dbname """
        new_dbname = dbname
        found_dbname = False
        extra_dbname = 1

        while not found_dbname:
            try:
                self.trial_check(new_dbname, spam_check=False) # no spam check to avoid infinite loop
                found_dbname = True
            except ValueError:
                # dbname already exists, find another one
                # ex : "myawesomedb1" -> "myawesomedb2"
                new_dbname = dbname + str(extra_dbname)
                extra_dbname += 1

        return new_dbname

    @api.model
    def select_host(self, hosting, version=None):
        """
        Select the hosting server.

        Selection is done is the following order:
        1. exact match on server name
        2. any "new" from chosen hosting location
        3. any "trial" from chosen hosting location
        4. first "new" from any hosting location
        5. first "trial" from any hosting location
        """
        base_domain = []
        if version:
            version = version.replace('~', '-')
            e_match = re.search(r'\+?e$', version)
            if e_match:
                version = version[:-len(e_match.group())]
            base_domain.extend([('last_branch', '=', version)])

        servers = self.sudo().search(base_domain + [('mode', 'like', 'trial'), ('name', '=like', hosting)])
        if not servers:
            servers = self.sudo().search(base_domain + [('mode', 'like', 'new'), ('hosting', '=ilike', hosting)])

        if not servers:
            servers = self.sudo().search(base_domain + [('mode', '=', 'trial'), ('hosting', '=', hosting)])

        # fallback, ignoring slaves and custom servers, selecting 1 single server deterministically
        if not servers:
            # Next option: another new trial in any hosting (presumably running last series), ordered alphabetically
            servers = self.sudo().search(base_domain + [('mode', 'like', 'new')], order='hosting', limit=1)
            _logger.info('select_host: first new in any hosting location: hosting=%s version=%s', hosting, version)
        if not servers:
            # Final option: any server strictly trial in the same hosting (non slave, not new, not custom)
            servers = self.sudo().search(base_domain + [('mode', '=', 'trial')], order='name', limit=1)
            _logger.info('select_host: first trial in any hosting location: hosting=%s version=%s', hosting, version)

        # Sort server by stat_db_all, and take the first third
        servers = servers.sorted(key=lambda r: int(r.stat_db_all or 0))[:len(servers) // 3 or 1]
        server = random.choice(servers)
        return server

    @api.model
    def book_dns(self, dbname, hosting, version=None, spam_check=True):
        dbname = dbname.lower()
        self.trial_check(dbname, spam_check=spam_check)

        # pick host
        server = self.select_host(hosting, version)

        # generate main dns entry for new dbname
        hostname = '%s.%s' % (dbname, DOMAIN_SAAS)
        self._create_dns(server, hostname)

        return hostname, server.name

    @api.model
    def _hostof(self, dbname):
        dbname = dbname.lower()
        domain = [
            ('name', '=', '%s.%s' % (dbname, DOMAIN_SAAS)),
            ('type', '=', 'MX'),
        ]

        dns = self.env['saas.dns'].search(domain, limit=1)
        if dns:
            return dns.content
        return None

    @api.model
    def _create_dns(self, server, hostname, origin='saas_booking'):
        """ Create DNS entries on the given server for the hostname
            :param server : saas.server record
            :param hostname : name of the instance (string)
        """
        values = {
            'origin': origin,
            'name': hostname,
            'type': 'A',
            'content': server.public_ip,
        }
        self.env['saas.dns'].create(values)
        # create mx
        values = {
            'origin': origin,
            'name': hostname,
            'type': 'MX',
            'content': server.name,
            'prio': 10
        }
        self.env['saas.dns'].create(values)
        # create spf
        values = {
            'origin': origin,
            'name': hostname,
            'type': 'TXT',
            'content': 'v=spf1 include:_spf.odoo.com -all',
        }
        self.env['saas.dns'].create(values)

    @api.model
    def _ensure_dns(self, server, hostname, origin, check_exists=True):
        records = self.env['saas.dns'].search([('name', '=', hostname)])
        if records:
            if not all(record.origin == origin for record in records):
                raise ValueError('This domain is already taken')
            if server:
                for record in records:
                    if record.type == 'A':
                        if not record.content == server.public_ip:
                            record.content = server.public_ip
                    elif record.type == 'MX':
                        if not record.content == server.name:
                            record.content = server.name
            else:
                records.unlink()
        else:
            if server:
                self.domain_check(hostname, origin, check_exists=check_exists)
                self._create_dns(server, hostname, origin)

    @api.model
    def _dup(self, from_, to, **kwargs):
        """ Duplicate the DNS entry of a given database name """
        from_ = from_.lower()
        to = to.lower()

        _logger.info('_dup: from: %r -> to: %r, flags: %r', from_, to, kwargs)

        if str2bool(kwargs.get('bypass_name_check', 'no')) is not True:
            self.trial_check(to)

        # search server for old
        hostname = '%s.%s' % (from_, DOMAIN_SAAS)
        dns = self.env['saas.dns'].search([('origin', '=ilike', 'saas%'), ('type', '=', 'A'), ('name', '=', hostname)], limit=1)
        if not dns:
            raise ValueError('_dup: no saas hostname found for %r' % from_)

        server = self.search([('public_ip', '=', dns.content), ('mode', 'like', '%trial%')], limit=1)
        if not server:
            raise LookupError('_dup: no trial server found for DNS IP %s' % dns.content)

        dest_hostname = '%s.%s' % (to, DOMAIN_SAAS)
        self._create_dns(server, dest_hostname)

        # if we want to drop the old dns
        if str2bool(kwargs.get('drop_old', 'no')) is True:
            # we need to remove all the DNS types (A, MX), so we need to fetch all of them
            dns = self.env['saas.dns'].search([('origin', '=ilike', 'saas%'), ('name', '=', hostname)])
            # and remove them
            dns.with_context(reason='renamed database, drop old one').unlink()

        return {'url': dest_hostname}

    @api.model
    def _dump_get_dbex_host(self, saas_server_name):
        """
        Find a suitable dbex host for the given trial server
        :param saas_server_name: name of the trial server
        :return: str host name
        """
        saas_server = self.search([('name', '=', saas_server_name)])
        if not saas_server:
            raise UserError('unknown saas server: {}'.format(saas_server_name))
        hosting = saas_server.hosting
        servers = self.search([('mode', 'like', 'dbex'), ('hosting', '=', hosting)])
        if not servers:
            servers = self.search([('mode', 'like', 'dbex')])
        if not servers:
            raise UserError('no candidate dbex host found for hosting: {}'.format(hosting))

        server = random.choice(servers)
        _logger.info('_dump_get_dbex_host: %s selected as host', server.name)
        return server.name

    @api.model
    def _restore_get_host(self, branch, source_db="", hosting=None, size_gb=0):
        """
        Finds suitable host for the given version
        :param branch: db version, branch format
        :param source_db: the name of the db, can be empty if hosting is given
        :param hosting: target hosting
        :param size_gb:
        :return:
        """
        host = self._hostof(source_db)
        if not host and not hosting:
            raise UserError('Unknown database: %s' % (source_db,))

        active_branches = set(self.search([('mode', 'like', 'trial')]).mapped('last_branch'))
        # retrieve the branch name from a version
        branch = humanize_version(branch)

        if branch not in active_branches:  # unknown backup version, default to original version
            _logger.info('_restore_get_host: unmanaged branch %s', branch)
            if host:
                branch = self.search([('name', '=', host)]).last_branch
            else:
                raise UserError('unmanaged branch: hosting: %s' % (hosting,))

        hosting = hosting or self.search([('name', '=', host)]).hosting

        disk_free = max(100, size_gb)
        domain = [
            ('last_branch', '=', branch),
            ('stat_disk_free', '>=', disk_free),
            ('mode', 'in', ('trial', 'trial new', 'trial test', 'trial test new')),
        ]

        # First, find a trial server with exact hosting match
        _logger.info(
            '_restore_get_host: looking for host candidates for %s: branch: %s, stat_disk_free: %s, hosting: %s',
            source_db,
            branch,
            disk_free,
            hosting,
        )
        servers = self.search(domain + [('hosting', '=', hosting)])

        # Then, find a trial server in the same region (exclude edu and dedicated servers)
        # Only applies to edu hosting
        if not servers and 'edu' in hosting and hosting != hosting[:2]:
            _logger.info(
                '_restore_get_host: looking for host candidates for %s: branch: %s, stat_disk_free: %s, hosting: %s',
                source_db,
                branch,
                disk_free,
                hosting[:2],
            )
            servers = self.search(domain + [('hosting', '=', hosting[:2])])

        # Do not fall back on other hosting, otherwise we may end up in the wrong region
        # because no server has enough available disk in the requested hosting.

        # Exception
        if not servers:
            raise UserError('no candidate restore host found for %s on branch: %s; hosting: %s' % (source_db, branch, hosting))

        # Sort server by stat_db_pay, and take the first third
        servers = servers.sorted(key=lambda r: int(r.stat_db_pay or 0))[:len(servers) // 3 or 1]
        server = random.choice(servers)
        _logger.info('_restore_get_host: %s selected as host for %s', server.name, source_db)
        return server.name

    @api.model
    def _restore_get_host_bak(self, branch, bak_server_name, source_db="", size_gb=0):
        """
        Finds suitable host for the given version based on the trial servers in the same
        backup group of the given backup server. The selection is based on the hosting of
        one first trial server from backup group.

        :param str branch: db version, branch format
        :param str bak_server_name: name of the backup server
        :param int size_gb: size of database

        :return str: a trial host matching the criteria
        """
        # We do not want to restore on specific hosting such as edu or dedicated servers
        hosting = None

        # First, try to find the hosting of the source db
        if source_db:
            _logger.info(
                "_restore_get_host_bak: looking for hosting based on source_db %s",
                source_db,
            )
            host = self._hostof(source_db)
            if host:
                hosting = self.search([
                    ("name", '=', host),
                    ("hosting", "!=", "reserved"),
                ]).hosting

        # If we still don't have a hosting, try to find the hosting of the backup server
        if not hosting:
            _logger.info(
                "_restore_get_host_bak: looking for hosting based on backup server %s",
                bak_server_name,
            )
            backup_server = self.search([("name", "=", bak_server_name)])
            if not backup_server:
                raise Exception("Unknown backup server: %s" % (bak_server_name,))
            if not backup_server.backup_group_id:
                raise Exception("Backup server %s has no backup group" % (bak_server_name,))
            hosting = self.search([
                ("backup_group_id", "=", backup_server.backup_group_id.id),
                ("mode", "like", "trial"),
                ("hosting", "!=", "reserved"),
            ], limit=1).hosting

        # If we still don't have a hosting, fall back to EU
        if not hosting:
            _logger.info("_restore_get_host_bak: fallback to EU hosting")
            hosting = "eu"
        return self._restore_get_host(branch, hosting=hosting, source_db=source_db, size_gb=size_gb)

    @api.model
    def _get_jail_host(self, branch, hosting=None):
        """
        Finds suitable jail host for the given version
        :param branch: db version, branch format
        :param hosting: target hosting
        :return: str host name
        """
        branch = humanize_version(branch)
        jail_servers = self.search([('mode', 'like', 'jail'), ('last_branch', '=', branch)])

        _logger.info('_get_jail_host: looking for host candidates for hosting: %s branch: %s', hosting, branch)

        if not jail_servers:
            raise Exception('no candidate jail host found on branch: %s' % branch)

        servers = jail_servers.filtered_domain([('hosting', '=', hosting)])
        # exceptions
        if not servers:
            _logger.info('no candidate jail host found for dump on branch: %s; hosting: %s, switching to other hosting',
                         branch, hosting)
            servers = jail_servers

        server = random.choice(servers)
        _logger.info('_get_jail_host: %s selected as host for %s', server.name, branch)
        return server.name

    @api.model
    def _get_sfu_host(self, hosting):
        """
        Finds a suitable SFU host for the given hosting. If no SFU is found for the given
        hosting, we fall back on EU servers, and finally on a random server.

        :param str hosting: hosting where the SFU should be located

        :return str: a sfu host matching the criteria
        """

        hosting = (hosting or "").lower()
        _logger.info("_get_sfu_host: looking for sfu host for hosting %s", hosting)
        servers = self.env["saas.server"].search([
            ("mode", "=", "sfu"),
            ("hosting", "=", hosting),
        ])

        # Fall back to EU server
        if not servers and hosting != "eu":
            _logger.warning("_get_sfu_host: no sfu host for hosting %s, fall back to eu", hosting)
            servers = self.env["saas.server"].search([
                ("mode", "=", "sfu"),
                ("hosting", "=", "eu"),
            ])

        # The following should never happen
        if not servers:
            _logger.error("_get_sfu_host: no sfu host found for hosting eu, fall back to random")
            servers = self.env["saas.server"].search([
                ("mode", "=", "sfu"),
            ])

        server = random.choice(servers)
        return server.name

    # -------------------------------------------------
    # Inter servers ping and server actions
    # -------------------------------------------------

    @api.model
    def _get_server(self, public_ip):
        """
        Return the server from the public IP. If no server is found, raise an exception.

        :param str public_ip: the IPv4 public address
        """

        domain = [('public_ip', '=', public_ip)]
        server = self.search(domain, limit=1)
        if not server:
            raise Exception("Host with IP {} not found".format(public_ip))
        return server

    @api.model
    def _dns_update(self, server, info):
        """
        Update DNS entries.
        For trial servers, update the origin of booked DNS entries, then garbage collect the old
        ones.
        For paas servers, garbage collect only.
        The following keys are relevant:
            hosts: list of domains hosted by the server
        """

        all_hosts = info.get('hosts', [])
        if isinstance(all_hosts, dict):
            all_hosts = []      # old version of custom ... ignore for now

        # cleanup dns entries...
        if all_hosts and 'trial' in (server.mode or '').split():
            # Update origin of booked DNS of the database
            self.env['saas.dns'].search([
                ('origin', 'in', ('saas_booking', 'saas_disaster')),
                ('name', 'in', all_hosts)
            ]).write({'origin': 'saas'})
            # Remove booked dns of this server older than two hours
            twohoursago = (datetime.datetime.now() + datetime.timedelta(hours=-2)).strftime('%Y-%m-%d %H:%M:%S')
            dns = self.env['saas.dns'].search([
                ('origin', '=', 'saas_booking'),
                ('create_date', '<', twohoursago),
                ('content', 'in', [server.public_ip, server.name])
            ])
            dns.with_context(reason='expired saas booking').unlink()
            # gc expired free trial and saas domains
            self.env.cr.execute("""
                SELECT array_agg(id) FROM records LEFT JOIN
                (SELECT id FROM records WHERE origin = 'saas' AND name = ANY(%s)) valid_records USING (id)
                WHERE content = ANY(%s) AND
                      origin = 'saas' AND
                      valid_records.id IS NULL;
            """, (all_hosts, [server.public_ip, server.name]))
            dns_ids = self.env.cr.fetchone()[0]
            dns = self.env['saas.dns'].browse(dns_ids)
            _logger.debug('Stat Update from %s - matched %s records', server.name, len(dns))

            dns.with_context(reason='dropped database').unlink()

        if all_hosts and 'paas_master' in (server.mode or '').split():
            # Raw SQL to avoid pathological query due to the number of entries in `all_hosts` (15s+)
            self.env.cr.execute("""
                SELECT array_agg(id) FROM records LEFT JOIN
                (SELECT id FROM records WHERE origin = 'paas' AND name = ANY(%s)) valid_records USING (id)
                WHERE create_date < now() - interval '2 hours' AND
                      origin = 'paas' AND
                      valid_records.id IS NULL;
            """, (all_hosts,))
            domain_ids = self.env.cr.fetchone()[0]
            domains = self.env['saas.dns'].browse(domain_ids)
            domains.with_context(reason='domain no longer used by the paas master').unlink()


    @api.model
    def _stats_update(self, server, info):
        """
        Update the statistics of the given server. The following keys are relevant:
            cpu
            disk_free
            db_count
            db_trial
            db_paid
            disk_total
            last_branch
            ram_total
            stat_loadavg
            stat_revs
        """

        # Update server stats
        values = {
            'stat_cpu': info.get('cpu', 0),
            'stat_db_all': info.get('db_count', 0),
            'stat_db_trial': info.get('db_trial', 0),
            'stat_db_pay': info.get('db_paid', 0),
            'stat_disk_free': info.get('disk_free', 0),
            'stat_disk_total': info.get('disk_total', 0),
            'stat_ram': info.get('ram_total', 0),
            'last_branch': info.get('last_branch', ''),
            'lastping': fields.datetime.now(),
        }

        for asit in 'loadavg revs'.split():
            values['stat_' + asit] = info.get(asit, '')

        server.write(values)

    @api.model
    @lru_cache(maxsize=1)
    def _saas_config(self, extra_fields=tuple(), cache_key=None):
        """
        Compute the list of worker objects for consumption by the stats and
        config endpoints.

        :param tuple extra_fields: extra fields to include in the return value.
        :param tuple cache_key: key used for cache invalidation.
        :return dict: the list of saas worker objects.
        """
        _logger.info("Computing all worker stats (cache_key=%s)", cache_key)

        # prepare pong
        pong_fields = "hosting lastping mode master name private_ip profiles public_dns public_ip public_ipv6 public_ssh_key http_shared_secret backuped last_branch cname_list".split()
        pong_fields += list(extra_fields)
        workers_read = self.search_read([], fields=pong_fields)
        workers = {}
        for i in workers_read:
            if not i['mode']:
                i['mode'] = ''      # no False
            if not i['profiles']:
                i['profiles'] = ''  # no False
            server = self.browse(i["id"])
            i["backup_group"] = server.backup_group_id.name if server.backup_group_id else ""
            i["dc"] = server.dc_id.name
            i["provider"] = server.dc_id.provider
            i['lastping'] = fields.Datetime.to_string(i['lastping'])
            i['replicas'] = sorted(server.replicas.mapped('name'))
            i['primaries'] = sorted(server.primaries.mapped('name'))
            del i['id']
            workers[i['name']] = i
        return workers

    @staticmethod
    def _trial_url(
        dbname, module, client_id, lang='en_US', template=None, hostname=None, odoo_com_uid=None,
        company_name=None, email=None, country_code=None, sync_after_creation=False, username=None, **_kw
    ):
        if LOCAL or not hostname:
            hostname = '%s.%s' % (dbname, SAAS_BASE_URL)
        scheme = SCHEME

        redirect_url = "{scheme}://{hostname}/saas_worker/trial_do?".format(**locals())

        encoded_params = werkzeug.urls.url_encode({
            'lang': lang,
            'module': module,
            'template': template,
            'dbname': dbname,
            'client_id': client_id,
            'country_code': country_code,
            'odoo_com_uid': odoo_com_uid,
            'company_name': company_name,
            'email': email,
            'sync_after_creation': sync_after_creation,
            'username': username,
            'confirm_token': _kw.get('confirm_token')
        })

        redirect_url = redirect_url + encoded_params
        thanks_url = "%s/thanks/trial?" % WEBSITE
        params = werkzeug.urls.url_encode(dict(app=module, dbname=dbname, re=redirect_url))

        return thanks_url + params

    # -------------------------------------------------
    # Provisioning
    # -------------------------------------------------

    def _set_backup_group(self, backup_group=None, disable_backup=False):
        """
        Assign or unassign a backup group to/from a server with the following logic:
        - Use the provided backup_group
        - Unassign any existing backup group if disable_backup is True
        - If no backup_group is passed and disabled_backup is False, fall back
          to the backup group auto assign feature

        :param str: backup_group
        :param bool: disable_backup
        """
        self.ensure_one()
        self = self.sudo()

        vals = {}
        if backup_group:
            bg = self.env['backup.group'].sudo().search([('name', '=', backup_group.lower())])
            if not bg:
                raise ValueError('Backup group %s does not exist' % backup_group.lower())
            vals["backup_group_id"] = bg.id
        elif disable_backup:
            vals["backup_group_id"] = None

        self.write(vals)

        # Auto-set backup group
        if not disable_backup and not self.backup_group_id:
            _logger.debug("Attempt to assign a backup group automatically")
            self._auto_set_backup_group()
            if not self.backup_group_id:
                raise ValueError("Could not assign backup group to server automatically.")

    def _server_delete_allowed(self):
        """
        Verify if a server can be deleted. The following checks are done:
        - no paid or trial DB is known
        - no MX record exist

        It raises if a condition is not met.
        """

        Dns = self.env["saas.dns"].with_user(SUPERUSER_ID)
        for server in self:
            if int(server.stat_db_pay) or int(server.stat_db_trial):
                raise UserError("Customer databases found on {}!".format(server.name))
            if Dns.search_count([("content", "=", server.name), ("type", "=", "MX")]):
                raise UserError("MX records found on {}!".format(server.name))

    def _make_spare(self):
        """
        Reset the server data in order to set an OVH server as spare. We make sure the server does
        not have any live DB running before reset.

        Advance2-only are supported.
        """

        self._server_delete_allowed()
        for server in self:
            if server.dc_id.provider != "OVH":
                raise UserError("Server {} is not from a supported provider".format(server.name))
            server.write({
                "name": "adv2-{}.odoo.com".format(binascii.hexlify(os.urandom(2)).decode()),
                "mode": "spare",
                "master": False,
                "backuped": False,
                "backup_group_id": False,
                "hosting": "am" if "BHS" in server.dc_id.name else "eu",
                "lastping": False,
                "last_branch": False,
                "stat_revs": False,
                "replicas": [(5, 0, 0)],
                "primaries": [(5, 0, 0)],
            })

    # -------------------------------------------------
    # Auto configuration
    # -------------------------------------------------

    def _auto_set_backup_group(self):
        """
        Set a backup group on a saas server based on the matching modes.

        A backup group is selected based on the following criteria:
        1. It must be available for auto_config
        2. It must match the server region
        3. The server and backup group modes must match

        We select the backup group with the highest number of matching modes. A backup group with
        no mode set is used as a fallback. A backup group with mode set but no match with the
        server won't be selected.
        """

        for server in self:
            if server.backup_group_id:
                _logger.warning("Backup group already set on %s", server.name)
                continue

            server_modes = set((server.mode or "").split())
            if not server_modes:
                _logger.warning("No mode set on server %s", server.name)
                continue

            # Get all available backup groups for the server region
            # Exceptions:
            # - AU servers are backed up on AS backup groups.
            # - SA servers are backed up on ME backup groups.
            region = server.name[:2]
            if region == "au":
                region = "as"
            elif region == "sa":
                region = "me"
            backup_groups = self.env["backup.group"].search(
                [
                    ("auto_config", "=", True),
                    ("name", "ilike", f"%{region}%"),
                ]
            )
            if not backup_groups:
                _logger.warning("Could not find any available backup group for region %s", region)
                continue

            # Select all matching backup groups
            bg_modes_match = defaultdict(list)
            for bg in backup_groups:
                bg_modes = set((bg.mode or "").split())
                match_modes = len(server_modes & bg_modes)

                # In case there are groups defined on the BG but no match with the server,
                # do not consider it as a possible match.
                if bg_modes and not match_modes:
                    continue

                bg_modes_match[match_modes].append(bg)

            if not bg_modes_match:
                _logger.warning("No available backup group match the server modes %s", server.mode)
                continue

            # Pick the backup group with the highest match of modes
            for k in sorted(bg_modes_match, reverse=True):
                # If multiple backup groups match, select the BG with the lowest number of nodes
                if len(bg_modes_match[k]) > 1:
                    server.backup_group_id = self.env["saas.server"].read_group(
                        domain=[("backup_group_id", "in", [s.id for s in bg_modes_match[k]])],
                        fields=["server_count:count(id)"],
                        groupby=["backup_group_id"],
                        limit=1,
                        orderby="server_count",
                    )[0]["backup_group_id"][0]
                else:
                    server.backup_group_id = bg_modes_match[k][0]
                break

    # -------------------------------------------------
    # Actions
    # -------------------------------------------------

    def action_open_server_rw(self):
        action = {
            "name": "Edit SaaS Server",
            "type": "ir.actions.act_window",
            "res_model": "saas.server",
            "views": [[self.env.ref('saas_master.saas_server_rw_form').id, "form"]],
            "res_id": self.id,
        }
        return action
